#!/usr/bin/env python

from kivy.logger import Logger
log = Logger.getChild("KiTT")

import gtk
import sys, dbus, subprocess, os

import wnck

from actions import Actions

def workspace_up():
    while gtk.events_pending():
        gtk.main_iteration()
    self.screen.get_workspace_neighbor(self.screen.get_active_workspace(), wnck.MOTION_LEFT).activate(0)
    return True

def workspace_down():
    while gtk.events_pending():
        gtk.main_iteration()
    self.screen.get_workspace_neighbor(self.screen.get_active_workspace(), wnck.MOTION_RIGHT).activate(0)
    return True

ACTIONS = dict(
    workspace_up = workspace_up,
    workspace_down = workspace_down
)




#!/usr/bin/env python

from kivy.logger import Logger
log = Logger.getChild("KiTT")

from kivy.event import EventDispatcher
from kivy.graphics import Line
from kivy.gesture import Gesture, GestureDatabase

import pyudev

from actions import Actions

def make_gesture(name, point_list):
    """
    A simple helper function
    """
    g = Gesture()
    g.add_stroke(point_list)
    g.normalize()
    g.name = name
    return g

class Listener(EventDispatcher):
    """
    listener function that queries kivy for touch events, builds the gesture
    and dispatch it through the actions singleton.
    """
    def __init__(self, config, gestures, el, *args, **kwarg):
        """
        :param config: string containing the path to the action configuration
        :param gestures: string containing the path to the gestures configuration
        """
        super(EventDispatcher, self).__init__(*args, **kwarg)
        self._event_loop = el
        self._gdb = GestureDatabase()
        self._actions = Actions(config, gestures)
        self.update_devices()
        self._multitouches = []

    def update_devices(self):
        log.debug('update_devices()')
        context = pyudev.Context()
        for device in context.list_devices(subsystem='input', ID_INPUT_MOUSE=True):
            if device.sys_name.startswith('event'):
                if 'PRODUCT' in device.parent.keys():
                    self._actions.update_gestures(device.parent['PRODUCT'])
        for gest_n, gest_r in self._actions.get_gestures().iteritems():
            for g in gest_r:
                g = self._gdb.str_to_gesture(g)
                g.normalize()
                g.name = gest_n
                self._gdb.add_gesture(g)

    def on_touch_down(self, touch):
        """
        listening function executed at begining of touch event
        builds the gesture
        """
        self._multitouches.append(touch)
        touch.ud['line'] = Line(points=(touch.sx, touch.sy))
        return True

    def on_touch_move(self, touch):
        """
        listening function executed during touch event
        store points of the gesture
        """
        # store points of the touch movement
        try:
            touch.ud['line'].points += [touch.sx, touch.sy]
            return True
        except (KeyError), e:
            pass

    def on_touch_up(self, touch):
        """
        touch is over, display informations, and check if it matches some
        known gesture.
        """
        if len(self._multitouches) is 0:
            return True

        log.debug("multitouches: \t%d" % len(self._multitouches))
        gestures = map(lambda g: self._gdb.find(make_gesture('',zip(touch.ud['line'].points[::2],
                                            touch.ud['line'].points[1::2])), minscore=0.70), self._multitouches)

        if not self._actions.dispatch(gestures, self._gdb):
            for touch in self._multitouches:
                log.debug("Touch:\t%s" % self._gdb.gesture_to_str(make_gesture('',zip(touch.ud['line'].points[::2],
                                                                                      touch.ud['line'].points[1::2]))))

        self._multitouches = []

    def on_motion(self, etype, me):
        """
        dispatches motion events
        """
        if etype == "begin":
            self.on_touch_down(me)
        elif etype == "update":
            self.on_touch_move(me)
        elif etype == "end":
            self.on_touch_up(me)
        else:
            log.error("Receive unknown event of type '%r': %s" % (etype, me))

    def dispatch(self, ev_type, ev_action, ev):
        """
        dispatches motion events
        """
        if ev_type == "on_motion":
            self.on_motion(ev_action, ev)
        else:
            log.error("asking to dispatch unknown event: '%r': '%r'" % (ev_type, ev))




#!/usr/bin/env python

from kivy.logger import Logger
log = Logger.getChild("KiTT")

import gtk
import sys
import imp
import json
import os.path

class Actions():
    """
    This class loads all the actions and gesture, and does the binding between
    them. Then, when a gesture is found by the listener, the gesture gets dispatched.
    """
    def __init__(self, actions_config, gestures_config):
        """
        Build and load actions and gestures from files
        :param actions_config: string containing the path to the actions JSON file
        :param gestures_config: string containing the path to the gestures JSON file
        """
        self._actions = dict(pinch_in=[],
                             pinch_out=[],
                             two_swipe_up=[],
                             two_swipe_down=[],
                             two_swipe_left=[],
                             two_swipe_right=[],
                             three_swipe_up=[],
                             three_swipe_down=[],
                             three_swipe_left=[],
                             three_swipe_right=[],
                             four_swipe_up=[],
                             four_swipe_down=[],
                             four_swipe_left=[],
                             four_swipe_right=[])
        self._functions = dict()
        self._gestures = dict()
        self._gestures_config = gestures_config
        try:
            with open(os.path.expanduser(actions_config)) as config:
                config = json.load(config)
                for engine in config['engines']:
                    plugin = imp.load_source("kitt.plugin_%s" % engine,
                                            "%s/plugin_%s.py" % (os.path.dirname(__file__), engine))
                    self._functions.update(plugin.ACTIONS)
                actions = config["actions"]
                for gesture, act_l in self._actions.iteritems():
                    if gesture in actions.keys():
                        for action in actions[gesture]:
                            act_l.append(action)
                    else:
                        log.error("Unable to load gesture: '%s' unknown" % gesture)
        except IOError:
            log.debug("No configuration file found")

    def update_gestures(self, model):
        log.debug('update_gestures')
        with open(os.path.expanduser(self._gestures_config)) as config:
            gestures = json.load(config)
            if model in gestures.keys():
                self._gestures.update(gestures[model])

    def get_gestures(self):
        """
        :return: the gesture dict
        """
        return self._gestures

    def before(self):
        """
        safeguard function to be called before a touch is defined
        """
        pass

    def after(self):
        """
        safeguard function to be called after a touch is defined
        """
        pass

    def dispatch(self, gestures, gdb):
        """
        dispatches found gestures, matched using the gesture database into
        gesture events.
        :param gestures: list of live gestures
        :param gdb: kivy's gesture database object
        """
        d = u = l = r = 0
        for gesture in gestures:
            if gesture is None:
                log.warning("Undefined touch")
                continue
            gesture = gesture[1]
            if gesture:
                if   gesture.name == 'move_down':  d += 1
                elif gesture.name == 'move_up':    u += 1
                elif gesture.name == 'move_left':  l += 1
                elif gesture.name == 'move_right': r += 1
                else:
                    log.warn("Unknown gesture")
            else:
                log.info("gesture: \t%s" % gdb.gesture_to_str(gesture))
                for gest_n, gest_r in self._gestures.iteritems():
                    s = gest_n, "\t",
                    for g2 in gest_r:
                        g2 = gdb.str_to_gesture(g2)
                        g2.normalize()
                        s += g2.get_score(gesture),
                    log.debug(s)

        if   u is 4 and d == l == r == 0: gesture = "four_swipe_up"
        elif d is 4 and u == l == r == 0: gesture = "four_swipe_down"
        elif l is 4 and u == d == r == 0: gesture = "four_swipe_left"
        elif r is 4 and u == d == l == 0: gesture = "four_swipe_right"
        elif u is 3 and d == l == r == 0: gesture = "three_swipe_up"
        elif d is 3 and u == l == r == 0: gesture = "three_swipe_down"
        elif l is 3 and u == d == r == 0: gesture = "three_swipe_left"
        elif r is 3 and u == d == l == 0: gesture = "three_swipe_right"
        elif u is 2 and d == l == r == 0: gesture = "two_swipe_up"
        elif d is 2 and u == l == r == 0: gesture = "two_swipe_down"
        elif l is 2 and u == d == r == 0: gesture = "two_swipe_left"
        elif r is 2 and u == d == l == 0: gesture = "two_swipe_right"
        else:
            log.warn("Gesture:\tNot found")
            return False

        for act in self._actions[gesture]:
            fun = act["function"]
            prm = act["parameters"]

            if fun in self._functions.keys():
                self._functions[fun](**prm)
            else:
                log.error("Action not found: %s" % fun)

        return True




#!/usr/bin/env python

from kivy.logger import Logger
log = Logger.getChild("KiTT")

from Xlib import X, XK, protocol, display, Xcursorfont
from Xlib.ext import xtest
from Xlib.protocol import request

import os

disp = display.Display()

def grab_mouse():
    log.debug("grab_mouse")

    cursor_font = disp.open_font('cursor')
    cursor = 0
    mycursor = cursor_font.create_glyph_cursor(cursor_font, cursor, cursor+1, (0,0,0), (0xFFFF, 0xFFFF, 0xFFFF))
       # grab_pointer(owner_events, event_mask, pointer_mode, keyboard_mode, confine_to, cursor, time)
    disp.screen().root.grab_pointer(False,                                        # owner_events
                                    X.Button4MotionMask | X.Button5MotionMask,                # event_mask
                                    X.GrabModeAsync,                              # pointer_mode
                                    X.GrabModeAsync,                              # keyboard_mode
                                    0,                                            # confine_to
                                    mycursor,                                     # cursor
                                    X.CurrentTime)                                # time
    disp.sync()


def ungrab_mouse():
    log.debug("ungrab_mouse")

    disp.ungrab_pointer(X.CurrentTime)
    disp.sync()


def do_mouse_click(button, target=None):
    """
    executes mouse button click

    :param button: X.Button1, X.Button2 or X.Button3 depending on which button to be triggered
    :param target: list containing name(s) of the target window
    """
    log.debug("do_mouse_click")
    root = disp.screen().root
    pointer_info = request.QueryPointer(display = disp.display,
                                        window = root)
    root_xpos, root_ypos = (pointer_info._data['root_x'], pointer_info._data['root_y'])
    targetwindow = disp.get_input_focus().focus

    if isinstance(button, basestring):
        if button is "Button1":
            button = X.Button1
        elif button is "Button2":
            button = X.Button2
        elif button is "Button3":
            button = X.Button3
        else:
            return False
    elif not button in (X.Button1, X.Button2, X.Button3):
        return False

    if targetwindow.get_wm_name() is None and targetwindow.get_wm_class() is None:
        targetwindow = targetwindow.query_tree().parent
    ret = targetwindow.translate_coords(root, root_xpos, root_ypos)
    target_xpos = ret.x
    target_ypos = ret.y

    if target:
        for target in target:
            if target in targetwindow.get_wm_class():
                break
        else:
            return False

    myevent_press = protocol.event.ButtonPress(detail = button,
                                            root=root,
                                            root_x=root_xpos,
                                            root_y=root_ypos,
                                            window=targetwindow.id,
                                            event_x=target_xpos,
                                            event_y=target_ypos,
                                            same_screen=1,
                                            state=0,
                                            time=X.CurrentTime,
                                            child=0)
    myevent_release = protocol.event.ButtonRelease(detail = button,
                                                root=root,
                                                root_x=root_xpos,
                                                root_y=root_ypos,
                                                window=targetwindow.id,
                                                event_x=target_xpos,
                                                event_y=target_ypos,
                                                same_screen=1,
                                                state=0,
                                                time=X.CurrentTime,
                                                child=0)

    # use window instead of display (xobject/drawable.py:send_event)
    disp.send_event(X.InputFocus,
                myevent_press,
                event_mask=0,
                propagate=1)

    # use window instead of display (xobject/drawable.py:send_event)
    disp.send_event(X.InputFocus,
                myevent_release,
                event_mask=0,
                propagate=1)
    disp.sync()
    return True


def do_key_press(keys, target=None):
    """
    executes a keypress

    :param keys: argument tuple containing one or several modifier keys, and the key to press
    :param target: list containing name(s) of the target window

    the key to pass as a parameter has to be taken from Xlib.XK library
    """
    log.debug("do_key_press")
    root = disp.screen().root

    keys = map(lambda k: XK.string_to_keysym(k) if isinstance(k, basestring) else k,
               keys)
    for key in keys:
        if not key in XK.__dict__.values():
            return False

    pointer_info = request.QueryPointer(display = disp.display,
                                        window = root)
    root_xpos, root_ypos = (pointer_info._data['root_x'], pointer_info._data['root_y'])
    targetwindow = disp.get_input_focus().focus
    if targetwindow.get_wm_name() is None and targetwindow.get_wm_class() is None:
        targetwindow = targetwindow.query_tree().parent
    ret = targetwindow.translate_coords(root, root_xpos, root_ypos)
    target_xpos = ret.x
    target_ypos = ret.y

    if target and targetwindow.get_wm_class():
        for t in target:
            if t in targetwindow.get_wm_class():
                break
        else:
            log.info("Window '%s' not found in target(s) %s" % (targetwindow.get_wm_class(), target))
            return False

    def send_key(display, window, keycodes):
        '''Send a KeyPress and KeyRelease event'''
        if not type(keycodes) in (tuple, list):
            keycodes = (keycodes,)
        # send with modifier
        for keycode in keycodes:
            xtest.fake_input(window,
                             X.KeyPress,
                             display.keysym_to_keycode(keycode))
        for keycode in reversed(keycodes):
            xtest.fake_input(window,
                             X.KeyRelease,
                             display.keysym_to_keycode(keycode))
        display.sync()

    send_key(disp, root, keys)
    return True


def switch_workspace(direction):
    """
    switches workspace

    :param direction: +1 for next workspace, -1 for previous workspace.
    """
    log.debug("switch_workspace")
    screen = disp.screen()
    root   = screen.root

    if isinstance(direction, basestring):
        direction = int(direction)

    def get_property(disp, name):
        atom = disp.intern_atom(name)
        return disp.screen().root.get_full_property(atom, 0)

    def send_event(win, ctype, data, mask=None):
        """ Send a ClientMessage event to the root """
        data = (data+[0]*(5-len(data)))[:5]
        ev = protocol.event.ClientMessage(window=win, client_type=ctype, data=(32,(data)))

        if not mask:
            mask = (X.SubstructureRedirectMask|X.SubstructureNotifyMask)
        root.send_event(ev, event_mask=mask)
        disp.sync()

    cur_ws = get_property(disp, '_NET_CURRENT_DESKTOP').value[0]
    nb_ws = get_property(disp, '_NET_NUMBER_OF_DESKTOPS').value[0]

    # switch to previous desktop
    if cur_ws + direction < 0:
        cur_ws = nb_ws -1
    elif cur_ws + direction >= nb_ws:
        cur_ws = 0
    else:
        cur_ws = cur_ws + direction
    send_event(root,
               display.Display().intern_atom("_NET_CURRENT_DESKTOP"),
               [cur_ws, X.CurrentTime])
    return True


ACTIONS = dict(
    workspace  = switch_workspace,
    keypress   = do_key_press,
    mouseclick = do_mouse_click
)

if __name__ == "__main__":
    import time
    print "grab_mouse_wheel"
    grab_mouse_wheel()
    time.sleep(1)
    print "do_mouse_click"
    do_mouse_click(X.Button3)
    time.sleep(1)
    print "do_key_press"
    do_key_press((XK.XK_Alt_L, XK.XK_Tab))
    time.sleep(1)
    print "ungrab_mouse_wheel"
    ungrab_mouse_wheel()
    # time.sleep(1)
    # print "switch_workspace"
    # switch_workspace(+1)



#!/usr/bin/env python

from kivy.logger import Logger
log = Logger.getChild("KiTT")

import gtk
import sys, dbus, subprocess, os

from kitt.actions import Actions

def dbus_message(service, path, method, parameters):
    """
    sends a dbus message
    :param service: string containing the full service name (with dots)
    :param path: string containing the full path to the object (with slashes)
    :param method: string containing name of the method to call
    :param parameters: list of each parameter
    """
    service = interface = '%s' % (service,)
    session_bus = dbus.SessionBus()
    proxy = session_bus.get_object(service, path)
    obj = dbus.Interface(proxy, interface)
    getattr(method)(*parameters)

ACTIONS = dict(send_dbus=dbus_message)




#!/usr/bin/env python

import sys
import argparse

import threading

from daemon import Daemon

import pyudev
import pyudev.glib

class UdevListener(threading.Thread):
    def __init__(self, gobject_loop, el, listener, log):
        threading.Thread.__init__(self)
        self._gobject_loop = gobject_loop
        self._event_loop = el
        self._listener = listener
        self.log = log

    def device_added(self, observer, device):
        for inp in self._event_loop.input_providers:
            if device.sys_name == inp.device:
                self.log.debug("Device '%s'/'%s' added ; starting event loop" % (device.sys_name, device.parent['PRODUCT']))
                self._event_loop.start()
                self._listener.update_devices()

    def device_removed(self, observer, device):
        for inp in self._event_loop.input_providers:
            if device.sys_name == inp.device:
                self.log.debug("Device '%s' removed ; stopping event loop" % (device.sys_name,))
                self._event_loop.stop()

    def run(self):
        self.log.debug("Starting udev listener")
        context = pyudev.Context()
        monitor = pyudev.Monitor.from_netlink(context)
        observer = pyudev.glib.GUDevMonitorObserver(monitor)

        observer.connect('device-added', self.device_added)
        observer.connect('device-removed', self.device_removed)

        monitor.enable_receiving()

        self._gobject_loop()


def run():
    """
    main function that parses the cli arguments and starts the application
    """
    parser = argparse.ArgumentParser(prog=sys.argv[0],
                                     description="KiTT: Kivy Touch Tool")
    parser.add_argument("-v", "--verbose",
                        dest="verbose",
                        action="store_true",
                        help="Enable verbose output")
    parser.add_argument("-c", "--config",
                        dest="config",
                        action="store",
                        default="~/.kivy/kitt_config.json",
                        help="Select the gesture plugin")
    parser.add_argument("-g", "--gestures",
                        dest="gestures",
                        action="store",
                        default="~/.kivy/kitt_gestures.json",
                        help="Select the gesture plugin")
    parser.add_argument("-p", "--pid",
                        dest="pid",
                        action="store",
                        default="/tmp/kitt.pid",
                        help="Where to store PID file")

    subparsers = parser.add_subparsers(
                        help='Daemon commands',
                        dest='commands')

    def main(args):
        """
        subfunction that launches the application
        """
        sys.argv = sys.argv[0:1]
        from kivy.logger import Logger, logging
        log = Logger.getChild("KiTT")
        if args.verbose:
            Logger.setLevel(logging.DEBUG)
        else:
            Logger.setLevel(logging.ERROR)
        from kivy.base import EventLoop, runTouchApp, stopTouchApp
        from kivy.support import install_gobject_iteration
        from listener import Listener

        listener = Listener(args.config, args.gestures, EventLoop)
        EventLoop.add_event_listener(listener)
        UdevListener(install_gobject_iteration, EventLoop, listener, log).start()
        runTouchApp()

    class Main(Daemon):
        """
        Daemonification class
        """
        def run(self):
            main(self.args)

    def do_start(args):
        Main(args.pid, args=args).start()

    def do_stop(args):
        Main(args.pid, args=args).stop()

    def do_foreground(args):
        main(args)

    subparsers.add_parser('start',
                          help = "Start daemon").set_defaults(func=do_start)
    subparsers.add_parser('stop',
                          help = "Stop daemon").set_defaults(func=do_stop)
    subparsers.add_parser('foreground',
                          help = "Start in foreground").set_defaults(func=do_foreground)

    args = parser.parse_args(sys.argv[1:])
    args.func(args)

if __name__ == "__main__":
    run()




from setuptools import setup
import os
import sys


def read(*names):
    values = dict()
    for name in names:
        if os.path.isfile(name):
            value = open(name).read()
        else:
            value = ''
        values[name] = value
    return values

long_description = """

%(README.md)s

""" % read('README.md')

setup(name='KiTT',
      version='1.1',
      description="Kivy Touch Tool",
      long_description=long_description,
      # classifiers=[],
      keywords='multitouch gesture management tool',
      author='Bernard `Guyzmo` Pratz',
      author_email='kitt@m0g.net',
      url='http://m0g.net/kitt/',
      license='GPLv3',
      packages=['kitt'],
      zip_safe=False,
      data_files=[('config', ['etc/kitt_config.json',
                              'etc/kitt_gestures.json'])],
      install_requires=[
          'kivy',
          'python-xlib',
          'pyudev',
          #'PyGTK',
          'argparse',
          'setuptools',
      ],
      entry_points="""
      # -*- Entry points: -*-
      [console_scripts]
      kitt = kitt.kitt:run
      """,
      )

if "install" in sys.argv:
    print """
Kitt is now installed!

Please install wnck package if you plan on using the wnck plugin:

    `apt-get install python-wnck`

And don't forget to add kitt to your .xsession/.xinit file!

"""
